@@ -19,12 +19,12 @@
 
 #include <string.h>
 
-PGRN_FUNCTION_INFO_V1(pgroonga_match_script_jsonb);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_match_script_jsonb);
 
 /* v2 */
-PGRN_FUNCTION_INFO_V1(pgroonga_match_jsonb);
-PGRN_FUNCTION_INFO_V1(pgroonga_query_jsonb);
-PGRN_FUNCTION_INFO_V1(pgroonga_script_jsonb);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_match_jsonb);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_query_jsonb);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_script_jsonb);
 
 typedef struct
 {
@@ -95,6 +95,42 @@ PGrnJSONBIteratorTokenToString(JsonbIteratorToken token)
 	}
 }
 
+const char *
+PGrnJSONBValueTypeToString(enum jbvType type)
+{
+	switch (type)
+	{
+	case jbvNull:
+		return "null";
+	case jbvString:
+		return "string";
+	case jbvNumeric:
+		return "numeric";
+	case jbvBool:
+		return "bool";
+	case jbvArray:
+		return "array";
+	case jbvObject:
+		return "object";
+	case jbvBinary:
+		return "binary";
+#ifdef PGRN_HAVE_JSONB_DATETIME
+	case jbvDatetime:
+		return "datetime";
+#endif
+	default:
+		return "unknown";
+	}
+}
+
+Jsonb *
+PGrnJSONBParse(const char *jsonString)
+{
+	Datum jsonbDatum =
+		DirectFunctionCall1(jsonb_in, CStringGetDatum(jsonString));
+	return DatumGetJsonbP(jsonbDatum);
+}
+
 grn_obj *
 PGrnJSONBLookupValuesTable(Relation index,
 						   unsigned int nthAttribute,
@@ -104,7 +140,7 @@ PGrnJSONBLookupValuesTable(Relation index,
 
 	snprintf(name, sizeof(name),
 			 PGrnJSONValuesTableNameFormat,
-			 index->rd_node.relNode, nthAttribute);
+			 PGRN_RELATION_GET_LOCATOR_NUMBER(index), nthAttribute);
 	return PGrnLookup(name, errorLevel);
 }
 
@@ -117,7 +153,7 @@ PGrnJSONBLookupPathsTable(Relation index,
 
 	snprintf(name, sizeof(name),
 			 PGrnJSONPathsTableNameFormat,
-			 index->rd_node.relNode, nthAttribute);
+			 PGRN_RELATION_GET_LOCATOR_NUMBER(index), nthAttribute);
 	return PGrnLookup(name, errorLevel);
 }
 
@@ -130,7 +166,7 @@ PGrnJSONBLookupTypesTable(Relation index,
 
 	snprintf(name, sizeof(name),
 			 PGrnJSONTypesTableNameFormat,
-			 index->rd_node.relNode, nthAttribute);
+			 PGRN_RELATION_GET_LOCATOR_NUMBER(index), nthAttribute);
 	return PGrnLookup(name, errorLevel);
 }
 
@@ -143,7 +179,7 @@ PGrnJSONBLookupFullTextSearchLexicon(Relation index,
 
 	snprintf(name, sizeof(name),
 			 PGrnJSONValueLexiconNameFormat,
-			 "FullTextSearch", index->rd_node.relNode, nthAttribute);
+			 "FullTextSearch", PGRN_RELATION_GET_LOCATOR_NUMBER(index), nthAttribute);
 	return PGrnLookup(name, errorLevel);
 }
 
@@ -156,7 +192,7 @@ PGrnJSONBLookupStringLexicon(Relation index,
 
 	snprintf(name, sizeof(name),
 			 PGrnJSONValueLexiconNameFormat,
-			 "String", index->rd_node.relNode, nthAttribute);
+			 "String", PGRN_RELATION_GET_LOCATOR_NUMBER(index), nthAttribute);
 	return PGrnLookup(name, errorLevel);
 }
 
@@ -169,7 +205,7 @@ PGrnJSONBLookupNumberLexicon(Relation index,
 
 	snprintf(name, sizeof(name),
 			 PGrnJSONValueLexiconNameFormat,
-			 "Number", index->rd_node.relNode, nthAttribute);
+			 "Number", PGRN_RELATION_GET_LOCATOR_NUMBER(index), nthAttribute);
 	return PGrnLookup(name, errorLevel);
 }
 
@@ -182,7 +218,7 @@ PGrnJSONBLookupBooleanLexicon(Relation index,
 
 	snprintf(name, sizeof(name),
 			 PGrnJSONValueLexiconNameFormat,
-			 "Boolean", index->rd_node.relNode, nthAttribute);
+			 "Boolean", PGRN_RELATION_GET_LOCATOR_NUMBER(index), nthAttribute);
 	return PGrnLookup(name, errorLevel);
 }
 
@@ -195,7 +231,7 @@ PGrnJSONBLookupSizeLexicon(Relation index,
 
 	snprintf(name, sizeof(name),
 			 PGrnJSONValueLexiconNameFormat,
-			 "Size", index->rd_node.relNode, nthAttribute);
+			 "Size", PGRN_RELATION_GET_LOCATOR_NUMBER(index), nthAttribute);
 	return PGrnLookup(name, errorLevel);
 }
 
@@ -636,6 +672,8 @@ PGrnJSONBInsertValue(JsonbIterator **iter,
 					 JsonbValue *value,
 					 PGrnJSONBInsertData *data)
 {
+	const char *tag = "[jsonb][insert]";
+
 	switch (value->type)
 	{
 	case jbvNull:
@@ -671,9 +709,9 @@ PGrnJSONBInsertValue(JsonbIterator **iter,
 		break;
 #ifdef PGRN_HAVE_JSONB_DATETIME
 	case jbvDatetime:
-		ereport(ERROR,
-				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-				 errmsg("pgroonga: datetime value for jsonb isn't supported")));
+		PGrnCheckRC(GRN_FUNCTION_NOT_IMPLEMENTED,
+					"%s datetime value isn't supported",
+					tag);
 		break;
 #endif
 	case jbvArray:
@@ -691,6 +729,7 @@ PGrnJSONBInsertValue(JsonbIterator **iter,
 static void
 PGrnJSONBInsertContainer(JsonbIterator **iter, PGrnJSONBInsertData *data)
 {
+	const char *tag = "[jsonb][insert][container]";
 	JsonbIteratorToken token;
 	JsonbIteratorToken lastToken = WJB_DONE;
 	JsonbValue value;
@@ -774,10 +813,10 @@ PGrnJSONBInsertContainer(JsonbIterator **iter, PGrnJSONBInsertData *data)
 			break;
 		}
 		default:
-			ereport(ERROR,
-					(errcode(ERRCODE_SYSTEM_ERROR),
-					 errmsg("pgroonga: jsonb iterator returns invalid token: %s",
-							PGrnJSONBIteratorTokenToString(token))));
+			PGrnCheckRC(GRN_UNKNOWN_ERROR,
+						"%s iterator returns invalid token: %s",
+						tag,
+						PGrnJSONBIteratorTokenToString(token));
 			break;
 		}
 	}
@@ -826,6 +865,7 @@ static void
 PGrnJSONBInsertContainerForFullTextSearch(JsonbIterator **iter,
 										  PGrnJSONBInsertData *data)
 {
+	const char *tag = "[jsonb][insert][container][full-text-search]";
 	JsonbIteratorToken token;
 	JsonbValue value;
 
@@ -844,10 +884,10 @@ PGrnJSONBInsertContainerForFullTextSearch(JsonbIterator **iter,
 		case WJB_END_OBJECT:
 			break;
 		default:
-			ereport(ERROR,
-					(errcode(ERRCODE_SYSTEM_ERROR),
-					 errmsg("pgroonga: jsonb iterator returns invalid token: %s",
-							PGrnJSONBIteratorTokenToString(token))));
+			PGrnCheckRC(GRN_UNKNOWN_ERROR,
+						"%s iterator returns invalid token: %s",
+						tag,
+						PGrnJSONBIteratorTokenToString(token));
 			break;
 		}
 	}
@@ -884,7 +924,7 @@ PGrnJSONBCreateTables(PGrnCreateData *data,
 		char jsonPathsTableName[GRN_TABLE_MAX_KEY_SIZE];
 		snprintf(jsonPathsTableName, sizeof(jsonPathsTableName),
 				 PGrnJSONPathsTableNameFormat,
-				 data->relNode, data->i);
+				 data->relNumber, data->i);
 		jsonbData->pathsTable =
 			PGrnJSONBCreatePathsTable(data->index, jsonPathsTableName);
 		GRN_PTR_PUT(ctx, data->supplementaryTables, jsonbData->pathsTable);
@@ -894,7 +934,7 @@ PGrnJSONBCreateTables(PGrnCreateData *data,
 		char jsonTypesTableName[GRN_TABLE_MAX_KEY_SIZE];
 		snprintf(jsonTypesTableName, sizeof(jsonTypesTableName),
 				 PGrnJSONTypesTableNameFormat,
-				 data->relNode, data->i);
+				 data->relNumber, data->i);
 		jsonbData->typesTable =
 			PGrnJSONBCreateTypesTable(data->index, jsonTypesTableName);
 		GRN_PTR_PUT(ctx, data->supplementaryTables, jsonbData->typesTable);
@@ -904,7 +944,7 @@ PGrnJSONBCreateTables(PGrnCreateData *data,
 		char jsonValuesTableName[GRN_TABLE_MAX_KEY_SIZE];
 		snprintf(jsonValuesTableName, sizeof(jsonValuesTableName),
 				 PGrnJSONValuesTableNameFormat,
-				 data->relNode, data->i);
+				 data->relNumber, data->i);
 		jsonbData->valuesTable =
 			PGrnJSONBCreateValuesTable(data->index, jsonValuesTableName);
 		GRN_PTR_PUT(ctx, data->supplementaryTables, jsonbData->valuesTable);
@@ -920,29 +960,32 @@ PGrnJSONBCreateFullTextSearchIndexColumn(PGrnCreateData *data,
 	grn_obj *type;
 	grn_obj *lexicon;
 	grn_obj *tokenizer = NULL;
-	grn_obj *normalizer = NULL;
+	grn_obj *normalizers = NULL;
 	grn_obj *tokenFilters = NULL;
+	grn_column_flags indexFlags = GRN_OBJ_COLUMN_INDEX | GRN_OBJ_WITH_POSITION;
 
 	PGrnApplyOptionValues(data->index,
+						  data->i,
 						  PGRN_OPTION_USE_CASE_FULL_TEXT_SEARCH,
 						  &tokenizer, PGRN_DEFAULT_TOKENIZER,
-						  &normalizer, PGRN_DEFAULT_NORMALIZER,
+						  &normalizers, PGRN_DEFAULT_NORMALIZERS,
 						  &tokenFilters,
-						  &flags);
+						  &flags,
+						  &indexFlags);
 
 	if (!tokenizer)
 		return;
 
 	snprintf(lexiconName, sizeof(lexiconName),
 			 PGrnJSONValueLexiconNameFormat,
-			 "FullTextSearch", data->relNode, data->i);
+			 "FullTextSearch", data->relNumber, data->i);
 	type = grn_ctx_at(ctx, GRN_DB_SHORT_TEXT);
 	lexicon = PGrnCreateTable(data->index,
 							  lexiconName,
 							  flags,
 							  type,
 							  tokenizer,
-							  normalizer,
+							  normalizers,
 							  tokenFilters);
 	GRN_PTR_PUT(ctx, data->lexicons, lexicon);
 
@@ -965,7 +1008,7 @@ PGrnJSONBCreateIndexColumn(PGrnCreateData *data,
 
 	snprintf(lexiconName, sizeof(lexiconName),
 			 PGrnJSONValueLexiconNameFormat,
-			 typeName, data->relNode, data->i);
+			 typeName, data->relNumber, data->i);
 	lexicon = PGrnCreateTable(data->index,
 							  lexiconName,
 							  tableType,
@@ -1043,13 +1086,14 @@ PGrnJSONBIsForFullTextSearchOnly(Relation index, unsigned int nthAttribute)
 void
 PGrnJSONBCreate(PGrnCreateData *data)
 {
+	const char *tag = "[jsonb][create]";
+
 	if (data->desc->natts != 1)
 	{
-		ereport(ERROR,
-				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-				 errmsg("pgroonga: multicolumn index for jsonb "
-						"isn't supported: <%s>",
-						data->index->rd_rel->relname.data)));
+		PGrnCheckRC(GRN_FUNCTION_NOT_IMPLEMENTED,
+					"%s multicolumn index isn't supported: <%s>",
+					tag,
+					data->index->rd_rel->relname.data);
 	}
 
 	if (PGrnJSONBIsForFullTextSearchOnly(data->index, data->i))
@@ -1088,7 +1132,7 @@ PGrnJSONBValueSetSource(Relation index,
 	snprintf(indexName, sizeof(indexName),
 			 PGrnJSONValueLexiconNameFormat ".%s",
 			 typeName,
-			 index->rd_node.relNode,
+			 PGRN_RELATION_GET_LOCATOR_NUMBER(index),
 			 nthAttribute,
 			 PGrnIndexColumnName);
 	if (required)
@@ -1689,6 +1733,7 @@ PGrnSearchBuildConditionJSONContain(PGrnSearchData *data,
 									grn_obj *targetColumn,
 									Jsonb *jsonb)
 {
+	const char *tag = "[build-condition][json-contain]";
 	unsigned int nthCondition = 0;
 	grn_obj components;
 	JsonbIterator *iter;
@@ -1763,10 +1808,10 @@ PGrnSearchBuildConditionJSONContain(PGrnSearchData *data,
 		case WJB_END_OBJECT:
 			break;
 		default:
-			ereport(ERROR,
-					(errcode(ERRCODE_SYSTEM_ERROR),
-					 errmsg("pgroonga: jsonb iterator returns invalid token: %s",
-							PGrnJSONBIteratorTokenToString(token))));
+			PGrnCheckRC(GRN_UNKNOWN_ERROR,
+						"%s iterator returns invalid token: %s",
+						tag,
+						PGrnJSONBIteratorTokenToString(token));
 			break;
 		}
 	}
@@ -1779,6 +1824,7 @@ PGrnJSONBBuildSearchCondition(PGrnSearchData *data,
 							  ScanKey key,
 							  grn_obj *targetColumn)
 {
+	const char *tag = "[build-condition][jsonb]";
 	unsigned int nthAttribute = 0;
 	grn_obj *subFilter;
 
@@ -1843,10 +1889,10 @@ PGrnJSONBBuildSearchCondition(PGrnSearchData *data,
 											DatumGetJsonbP(key->sk_argument));
 		break;
 	default:
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("pgroonga: jsonb: unsupported strategy number: %d",
-						key->sk_strategy)));
+		PGrnCheckRC(GRN_INVALID_ARGUMENT,
+					"%s unexpected strategy number: %d",
+					tag,
+					key->sk_strategy);
 		break;
 	}
 	return true;
@@ -2043,33 +2089,33 @@ PGrnJSONBRemoveUnusedTables(Oid relationFileNodeID)
 			return;
 	}
 
-		PGrnRemoveJSONValueLexicon("FullTextSearch", relationFileNodeID);
-		PGrnRemoveJSONValueLexicon("String", relationFileNodeID);
-		PGrnRemoveJSONValueLexicon("Number", relationFileNodeID);
-		PGrnRemoveJSONValueLexicon("Boolean", relationFileNodeID);
-		PGrnRemoveJSONValueLexicon("Size", relationFileNodeID);
+	PGrnRemoveJSONValueLexicon("FullTextSearch", relationFileNodeID);
+	PGrnRemoveJSONValueLexicon("String", relationFileNodeID);
+	PGrnRemoveJSONValueLexicon("Number", relationFileNodeID);
+	PGrnRemoveJSONValueLexicon("Boolean", relationFileNodeID);
+	PGrnRemoveJSONValueLexicon("Size", relationFileNodeID);
 
-		{
-			char name[GRN_TABLE_MAX_KEY_SIZE];
+	{
+		char name[GRN_TABLE_MAX_KEY_SIZE];
 
-			snprintf(name, sizeof(name),
-					 PGrnJSONPathsTableNameFormat ".%s",
-					 relationFileNodeID, 0, PGrnIndexColumnName);
-			PGrnRemoveObject(name);
+		snprintf(name, sizeof(name),
+				 PGrnJSONPathsTableNameFormat ".%s",
+				 relationFileNodeID, 0, PGrnIndexColumnName);
+		PGrnRemoveObject(name);
 
-			snprintf(name, sizeof(name),
-					 PGrnJSONValuesTableNameFormat,
-					 relationFileNodeID, 0);
-			PGrnRemoveObject(name);
+		snprintf(name, sizeof(name),
+				 PGrnJSONValuesTableNameFormat,
+				 relationFileNodeID, 0);
+		PGrnRemoveObject(name);
 
-			snprintf(name, sizeof(name),
-					 PGrnJSONPathsTableNameFormat,
-					 relationFileNodeID, 0);
-			PGrnRemoveObject(name);
+		snprintf(name, sizeof(name),
+				 PGrnJSONPathsTableNameFormat,
+				 relationFileNodeID, 0);
+		PGrnRemoveObject(name);
 
-			snprintf(name, sizeof(name),
-					 PGrnJSONTypesTableNameFormat,
-					 relationFileNodeID, 0);
-			PGrnRemoveObject(name);
-		}
+		snprintf(name, sizeof(name),
+				 PGrnJSONTypesTableNameFormat,
+				 relationFileNodeID, 0);
+		PGrnRemoveObject(name);
+	}
 }
